;;  *** IMP NODE #4 SETUP ***

;   IMP #4 coneects to IMP #3 via modem line 1 on this end and modem line
; 2 on the IMP 3 end ...

; Set the simulator configuration ...
echo Creating standard configuration for IMP #4 ...
do impconfig.cmd
SET IMP NUM=4

; Load the IMP code ...
echo Loading IMP code ...
do impcode.cmd

; Start up the modem links!
echo Attaching modem links ...
ATTACH MI1 4441::4432

; And we're done ..
echo Type GO to start ...
