;;  *** IMP NODE #2 SETUP ***

;  IMP #2 connects to IMP #2 via modem line 1 on both ends ...

; Set the simulator configuration ...
echo Creating standard configuration for IMP #2 ...
do impconfig.cmd
SET IMP NUM=2

; Load the IMP code ...
echo Loading IMP code ...
do impcode.cmd

; Start up the modem links!
echo Attaching modem links ...
ATTACH MI1 4421::4431

; And we're done ..
echo Type GO to start ...
