{0x01, "DUP - COMMON SUBROUTINES"},
{0x02, "DUP - CTRL RECORD PROCESSOR"},
{0x03, "DUP - STORE PHASE"},
{0x04, "DUP - *FILES, *LOCAL, *NOCAL PHASE"},
{0x05, "DUP - DUMP PHASE"},
{0x06, "DUP - DUMP LET/FLET PHASE"},
{0x07, "DUP - DELETE PHASE"},
{0x08, "DUP - DEFINE PHASE"},
{0x09, "DUP - EXIT PHASE"},
{0x0A, "DUP - CARD I/O INTERFACE"},
{0x0B, "DUP - KEYBOARD INPUT INTERFACE"},
{0x0C, "DUP - PAPER TAPE I/O INTERFACE"},
{0x0D, "DUP - SAVED UPCOR PHASE"},
{0x0E, "DUP - PRINCIPAL I/O DUMMY PHASE"},
{0x0F, "DUP - PRINCIPAL I/O (W/O KB) DUMMY PHASE"},
{0x10, "DUP - PAPER TAPE I/O DUMMY PHASE"},
{0x11, "DUP - MOVE DCI PROGRAMS TO UA OR FXA"},
{0x12, "DUP - EXIT TO MODIF DUMMY PHASE"},

{0x1F, "FOR - INPUT PHASE"},
{0x20, "FOR - CLASSIFIERPHASE"},
{0x21, "FOR - CHECK ORDER/ST NO PHASE"},
{0x22, "FOR - COMMON SUBR OR FUNCTION PHASE"},
{0x23, "FOR - DIMENSION/REAL/INTEGER PHASE"},
{0x24, "FOR - REAL CONSTANT PHASE"},
{0x25, "FOR - DEFINE FILE, CALL LINK/EXIT PHASE"},
{0x26, "FOR - VARIABLE, STMT FUNC PHASE"},
{0x27, "FOR - DATA STATEMENT PHASE"},
{0x28, "FOR - FORMAT STATEMENT PHASE"},
{0x29, "FOR - SUBTRACT DECOMPOSITION PHASE"},
{0x2A, "FOR - ASCAN I PHASE"},
{0x2B, "FOR - ASCAN II PHASE"},
{0x2C, "FOR - DO/CONTINUE/ETC PHASE"},
{0x2D, "FOR - SUBSCRIPT OPTIMIZATION PHASE"},
{0x2E, "FOR - SCAN PHASE"},
{0x2F, "FOR - EXPANDER I PHASE"},
{0x30, "FOR - EXPANDER II PHASE"},
{0x31, "FOR - DATA ALLOCATION PHASE"},
{0x32, "FOR - COMPILATION ERROR PHASE"},
{0x33, "FOR - STATEMENT ALLOCATION PHASE"},
{0x34, "FOR - LIST STATEMENT PHASE"},
{0x35, "FOR - LIST SYMBOL TABLE PHASE"},
{0x36, "FOR - LIST CONSTANTS PHASE"},
{0x37, "FOR - OUTPUT I PHASE"},
{0x38, "FOR - OUTPUT II PHASE"},
{0x39, "FOR - RECOVERY/EXIT PHASE"},

{0X51, "COBOL 51"},
{0X52, "COBOL 52"},
{0X53, "COBOL 53"},
{0X54, "COBOL 54"},
{0X55, "COBOL 55"},
{0X56, "COBOL 56"},
{0X57, "COBOL 57"},
{0X58, "COBOL 58"},
{0X59, "COBOL 59"},
{0X5A, "COBOL 5A"},
{0X5B, "COBOL 5B"},
{0X5C, "COBOL 5C"},

{0X6E, "SUP PHASE 1 - MONITOR CONTROL RECORD ANALYZER"},
{0x6F, "SUP PHASE 2 - JOB PROCESSING"},
{0x70, "SUP PHASE 3 - DELETE TEMPORARY LET"},
{0x71, "SUP PHASE 4 - XEQ PROCESSING"},
{0x72, "SUP PHASE 5 - SUPV CONTROL REC PROCESSING"},
{0X73, "SYSTEM DUMP-CORE-TO-PRINTER"},
{0X74, "AUXILIARY SUPERVISOR"},
{0X78, "CORE LOAD BUILDER, PHASE 1"},
{0x79, "CORE LOAD BUILDER, PHASE 2"},
{0x7A, "CORE LOAD BUILDER, PHASE 3"},
{0x7B, "CORE LOAD BUILDER, PHASE 4"},
{0x7C, "CORE LOAD BUILDER, PHASE 5"},
{0x7D, "CORE LOAD BUILDER, PHASE 6"},
{0x7E, "CORE LOAD BUILDER, PHASE 7"},
{0x7F, "CORE LOAD BUILDER, PHASE 8"},
{0x80, "CORE LOAD BUILDER, PHASE 9"},
{0x81, "CORE LOAD BUILDER, PHASE 10"},
{0x82, "CORE LOAD BUILDER, PHASE 11"},
{0x83, "CORE LOAD BUILDER, PHASE 12"},
{0x84, "CORE LOAD BUILDER, PHASE 13"},

{0X8C, "SYS 1403 READER"},
{0x8D, "SYS 1132 PRINTER"},
{0x8E, "SYS CONSOLE PRINTER"},
{0x8F, "SYS 2501/1442 READER"},
{0x90, "SYS 1442/1442 READER"},
{0x91, "SYS 1134/1055 PAPER TAPE IO"},
{0x92, "SYS KEYBOARD"},
{0x93, "SYS 2501/1442 CONVERSION"},
{0x94, "SYS 1134/1055 CONVERSION"},
{0x95, "SYS KEYBOARD CONVERSION"},
{0x96, "DISKZ"},
{0x97, "SYS DISK1"},
{0x98, "SYS DISKN"},

{0xA0, "CIL CORE IMAGE LOADER - PHASE 1"},
{0xA1, "CIL CORE IMAGE LOADER - PHASE 2"},

{0XB0, "RPG B0"},
{0XB1, "RPG B1"},
{0XB2, "RPG B2"},
{0XB3, "RPG B3"},
{0XB4, "RPG B4"},
{0XB5, "RPG B5"},
{0XB6, "RPG B6"},
{0XB7, "RPG B7"},
{0XB8, "RPG B8"},
{0XB9, "RPG B9"},
{0XBA, "RPG BA"},
{0XBB, "RPG BB"},
{0XBC, "RPG BC"},
{0XBD, "RPG BD"},
{0XBE, "RPG BE"},
{0XBF, "RPG BF"},
{0XC0, "RPG C0"},
{0XC1, "RPG C1"},
{0XC2, "RPG C2"},
{0XC3, "RPG C3"},
{0XC4, "RPG C4"},
{0XC5, "RPG C5"},
{0XC6, "RPG C6"},
{0XC7, "RPG C7"},
{0XC8, "RPG C8"},
{0XC9, "RPG C9"},
{0XCA, "RPG CA"},
{0XCB, "RPG CB"},
{0XCC, "RPG CC"},

{0XCD, "DUP PART 2 - CTRL"},
{0XCE, "DUP PART 2 - MACRO UPDATE"},

{0XCF, "ASM INITIALIZATION PHASE"},
{0xD0, "ASM CARD CONVERSION PHASE"},
{0xD1, "ASM DSF OUTPUT PHASE"},
{0xD2, "ASM INTERMEDIATE INPUT PHASE"},
{0xD3, "ASM END STATEMENT PHASE"},
{0xD4, "ASM ASSEMBLY ERROR PHASE"},
{0xD5, "ASM CONTROL CARDS I"},
{0xD6, "ASM CONTROL CARDS 2"},
{0xD7, "ASM DUMMY SYST SYMBOL TBL"},
{0xD8, "ASM SYMBOL TABLE OPTIONS PHASE"},
{0xD9, "ASM EXIT PHASE"},
{0xDA, "ASM PROG HEADER MNEMONICS PH"},
{0xDB, "ASM FILE STATEMENT PHASE"},
{0xDC, "ASM COMMON SUBROUTINES,ASCOM"},
{0xE4, "ASM INTERMEDIATE I/O"},
{0xE5, "ASM SYMBOL TABLE OVERFLOW"},
{0xDD, "ASM PROG CONTROL MNEMONICS PH"},
{0xDE, "ASM IMPERATIVE STATEMENTS PH"},
{0xDF, "ASM DECML,XFLC PROCESSING PH"},
{0xE0, "ASM DECIMAL CONVERSION PH"},
{0xE1, "ASM PROG LINKAGE PHASE"},
{0xE2, "ASM DMES PROCESSING PHASE"},
{0xE3, "ASM PUNCH CONVERSION PHASE"},
{0xE6, "ASM GRAPHIC ORDER PHASE"},
{0xE8, "ASM CONTROL CARDS III"},
{0xE9, "ASM MACRO PH 1 - SPECIAL OP AND PREPROCESSI"},
{0xEA, "MACRO PHASE 1A - SPECIAL PSEUDO OPS"},
{0xEB, "MACRO PHASE 1B - CONDITIONAL ASM PSEUDO OPS"},
{0xEC, "ASM MACRO PHASE 2 - MACRO DEFINITION"},
{0xED, "MACRO PHASE 2A - MACRO DEFINITION"},
{0xEE, "MACRO PHASE 2B - MACRO DEFINITION"},
{0xEF, "MACRO PHASE 3 - MACRO EXPANSION"},
{0xF0, "MACRO PHASE 3A - MACRO EXPANSION"},
{0xF1, "MACRO PHASE 3B - MACRO EXPANSION"},
{0xE7, "ASM DIVISION OPERATOR"},
{0xF2, "ASM CROSS-REFERENCE PART I"},
{0xF3, "ASM CROSS-REFERENCE PART 2A"},
{0xF4, "ASM CROSS-REFERENCE PART 2B"},
{0xF5, "ASM CROSS-REFERENCE PART 2C"},
{0xF6, "ASM CROSS-REFERENCE PART III"},
